@@ -1,23 +1,13 @@
-/******************************** LICENSE ********************************
+/*
+ * (C) Copyright 1996-2016 ECMWF.
+ *
+ * This software is licensed under the terms of the Apache Licence Version 2.0
+ * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
+ * In applying this licence, ECMWF does not waive the privileges and immunities
+ * granted to it by virtue of its status as an intergovernmental organisation nor
+ * does it submit to any jurisdiction.
+ */
 
-
-  Copyright 2007 European Centre for Medium-Range Weather Forecasts (ECMWF)
-
-  Licensed under the Apache License, Version 2.0 (the "License");
-  you may not use this file except in compliance with the License.
-  You may obtain a copy of the License at
-
-     http://www.apache.org/licenses/LICENSE-2.0
-
-  Unless required by applicable law or agreed to in writing, software
-  distributed under the License is distributed on an "AS IS" BASIS,
-  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-  See the License for the specific language governing permissions and
-  limitations under the License.
-European Centre for Medium-Range Weather Forecasts
-
-
- ******************************** LICENSE ********************************/
 /*
 
  I currently have a function that accepts a cairo context (with a path already set)
@@ -42,7 +32,7 @@ Something like:
 
     Started: Mon Oct 15 20:49:32 2007
 
-   \todo Fix 'convert' dependency
+   \todo Fix 'convert' dependency in "renderImage"
    \todo Check how much drivers are dependent on writing temp files in local directory (thread safety)
 */
 #include <cairo.h>
@@ -82,16 +72,13 @@ Display *dpy;
 
 #define FONT_SCALE 25*.7  //! \todo clean-up!!!
 
-extern "C"{
-#include "libimagequant/pngquant.h"
-}
 
 using namespace magics;
 
 /*!
   \brief Constructor
 */
-CairoDriver::CairoDriver() : filename_(""), offsetX_(0), offsetY_(0), backend_("PDF")
+CairoDriver::CairoDriver() : offsetX_(0), offsetY_(0), backend_("PDF")
 {
         cr_ = 0;
 }
@@ -108,7 +95,7 @@ CairoDriver::~CairoDriver()
 */
 void CairoDriver::open()
 {
-	MagLog::dev() << "Cairo version used is: "<<cairo_version_string()<< " backend: "<<backend_ << std::endl;
+	MagLog::info() << "Cairo version used is: "<<cairo_version_string()<< " backend: "<<backend_ << std::endl;
 
 	MFloat ratio = getYDeviceLength() / getXDeviceLength();
 	int width = maground(width_);
@@ -146,7 +133,7 @@ void CairoDriver::open()
 	cairo_status_t res = cairo_surface_status(surface_);
 	if (res != CAIRO_STATUS_SUCCESS)
 	{
-		MagLog::warning() << "Cairo> " << cairo_status_to_string(res) << endl;
+		MagLog::warning() << "Cairo > " << cairo_status_to_string(res) << endl;
 		return;
 	}
 
@@ -168,11 +155,24 @@ void CairoDriver::setupNewSurface() const
 	else if(magCompare(backend_,"pdf"))
 	{
 #if CAIRO_HAS_PDF_SURFACE
-	    filename_ = getFileName("pdf");
-//            dimensionXglobal_ = static_cast<int>(getXDeviceLength()*72/2.54);
- //           dimensionYglobal_ = static_cast<int>(getYDeviceLength()*72/2.54);
+	    fileName_ = getFileName("pdf");
+	    surface_  = cairo_pdf_surface_create(fileName_.c_str(), dimensionXglobal_, dimensionYglobal_);
+
+#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 5, 0)
+/*
+  typedef enum _cairo_pdf_metadata {
+    CAIRO_PDF_METADATA_TITLE,
+    CAIRO_PDF_METADATA_AUTHOR,
+    CAIRO_PDF_METADATA_SUBJECT,
+    CAIRO_PDF_METADATA_KEYWORDS,
+    CAIRO_PDF_METADATA_CREATOR,
+    CAIRO_PDF_METADATA_CREATE_DATE,
+    CAIRO_PDF_METADATA_MOD_DATE,
+} cairo_pdf_metadata_t;
+*/
+// cairo_pdf_surface_set_metadata(CAIRO_PDF_METADATA_CREATOR, "Magics");
+#endif
 
-	    surface_ = cairo_pdf_surface_create(filename_.c_str(), dimensionXglobal_, dimensionYglobal_);
 #else
 	    MagLog::error() << "CairoDriver: PDF output NOT supported! Enable PDF support in your Cairo installation." << std::endl;
 #endif
@@ -180,19 +180,10 @@ void CairoDriver::setupNewSurface() const
 	else if(magCompare(backend_,"ps"))
 	{
 #if CAIRO_HAS_PS_SURFACE
-	    filename_ = getFileName("ps"); 
-
-	    const int dimensionXglobal = static_cast<int>(getXDeviceLength()*72/2.54);
-	    const int dimensionYglobal = static_cast<int>(getYDeviceLength()*72/2.54);
-
-//	    if(dimensionXglobal>dimensionYglobal)   // landscape
-//	    {
-//               surface_ = cairo_ps_surface_create(filename_.c_str(), dimensionYglobal,dimensionXglobal);
-//            }
-//            else
-//            {
-                surface_ = cairo_ps_surface_create(filename_.c_str(), dimensionXglobal,dimensionYglobal);
-//            }
+        fileName_ = getFileName("ps");
+        const int dimensionXglobal = static_cast<int>(getXDeviceLength()*72/2.54);
+        const int dimensionYglobal = static_cast<int>(getYDeviceLength()*72/2.54);
+        surface_ = cairo_ps_surface_create(fileName_.c_str(), dimensionXglobal,dimensionYglobal);
 #else
 	    MagLog::error() << "CairoDriver: PS output NOT supported! Enable PS support in your Cairo installation." << std::endl;
 #endif
@@ -200,56 +191,27 @@ void CairoDriver::setupNewSurface() const
 	else if(magCompare(backend_,"eps"))
 	{
 #if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 5, 2)
-	    filename_ = getFileName("eps");
-	    surface_ = cairo_ps_surface_create(filename_.c_str(), dimensionXglobal_, dimensionYglobal_);
+	    fileName_ = getFileName("eps");
+	    surface_ = cairo_ps_surface_create(fileName_.c_str(), dimensionXglobal_, dimensionYglobal_);
 	    cairo_ps_surface_set_eps (surface_,true);
 #else
 	    MagLog::error() << "CairoDriver: EPS output NOT supported! You need at least version Cairo 1.5.2.\n"
 	                 << "PostScript is generated instead." << std::endl;
-	    filename_ = getFileName("ps");
-	    surface_ = cairo_ps_surface_create(filename_.c_str(), dimensionXglobal_, dimensionYglobal_);
+	    fileName_ = getFileName("ps");
+	    surface_ = cairo_ps_surface_create(fileName_.c_str(), dimensionXglobal_, dimensionYglobal_);
 #endif
 	}
 	else if(magCompare(backend_,"svg"))
 	{
 #if CAIRO_HAS_SVG_SURFACE
-	    filename_ = getFileName("svg",currentPage_+1);
+	    fileName_ = getFileName("svg",currentPage_+1);
 
-	    surface_ = cairo_svg_surface_create (filename_.c_str(), dimensionXglobal_, dimensionYglobal_);
+	    surface_ = cairo_svg_surface_create (fileName_.c_str(), dimensionXglobal_, dimensionYglobal_);
 //	    cairo_svg_surface_restrict_to_version (surface_, CAIRO_SVG_VERSION_1_2);
 #else
 	    MagLog::error() << "CairoDriver: SVG output NOT supported! Enable SVG support in your Cairo installation." << std::endl;
 #endif
 	}
-/*
-	else if(magCompare(backend_,"x"))
-	{
-#if CAIRO_HAS_XLIB_SURFACE
-		Window rootwin;
-		Window win;
-
-		if(!(dpy=XOpenDisplay(NULL)))
-		{
-			MagLog::error() << "CairoDriver: Could not open display for Xlib!"<< std::endl;
-			terminate();
-		}
-
-		int scr=DefaultScreen(dpy);
-		rootwin=RootWindow(dpy, scr);
-
-		win=XCreateSimpleWindow(dpy, rootwin, 1, 1, dimensionXglobal_, dimensionYglobal_, 0,
-				BlackPixel(dpy, scr), BlackPixel(dpy, scr));
-
-		XStoreName(dpy, win, "Magics++");
-		XSelectInput(dpy, win, ExposureMask|ButtonPressMask);
-		XMapWindow(dpy, win);
-
-		surface_ = cairo_xlib_surface_create(dpy, win, DefaultVisual(dpy, 0), dimensionXglobal_, dimensionYglobal_);
-#else
-		MagLog::error() << "CairoDriver: Xlib output NOT supported! Enable Xlib support in your Cairo installation." << std::endl;
-#endif
-	}
-*/
 	else
 	{
 		MagLog::error() << "CairoDriver: The backend "<< backend_ <<" is NOT supported!" << std::endl;
@@ -258,7 +220,7 @@ void CairoDriver::setupNewSurface() const
 	cairo_status_t status = cairo_surface_status(surface_);
 	if (status)
 	{
-		MagLog::error()	<< "CairoDriver: the surface ("<<backend_<<") could not be generated!\n"
+		MagLog::error()	<< "CairoDriver: the output file ("<<backend_<<") could NOT be generated!"
 				<< " -> "<<cairo_status_to_string(status)<< std::endl;
 	}
 
@@ -271,37 +233,20 @@ void CairoDriver::setupNewSurface() const
 	  const SystemInfo info;
 	  const string s1 = "%%Title: " + title_;
 	  cairo_ps_surface_dsc_comment (surface_, s1.c_str());
-	  const string s2 = "%%Creator2: " + getMagicsVersionString();
+	  const string s2 = "%%Creator2: "+ output_creator_+ " "+ getMagicsVersionString();
 	  cairo_ps_surface_dsc_comment (surface_, s2.c_str());
 	  const string s3 = "%%For: " + info.getUserID() + "@" + info.getHostName() + " " + info.getUserName();
 	  cairo_ps_surface_dsc_comment (surface_, s3.c_str());
 
-/*	    if(dimensionXglobal_>dimensionYglobal_)   // landscape
-	    {
-                dimensionYglobal_ = static_cast<int>(getXDeviceLength()*72/2.54);
-                dimensionXglobal_ = static_cast<int>(getYDeviceLength()*72/2.54);
-                cairo_translate (cr_, 0, dimensionYglobal_);
-                cairo_matrix_t matrix;
-                cairo_matrix_init (&matrix, 0, -1, 1, 0, 0,  0);
-		cairo_transform (cr_, &matrix);
-		cairo_ps_surface_dsc_comment (surface_, "%%PageOrientation: Landscape");
-	    }
-            else
-            {
-*/
-                dimensionXglobal_ = static_cast<int>(getXDeviceLength()*72/2.54);
-                dimensionYglobal_ = static_cast<int>(getYDeviceLength()*72/2.54);
-          	cairo_ps_surface_dsc_comment (surface_, "%%PageOrientation: Landscape");
-//            }
+      dimensionXglobal_ = static_cast<int>(getXDeviceLength()*72/2.54);
+      dimensionYglobal_ = static_cast<int>(getYDeviceLength()*72/2.54);
+      cairo_ps_surface_dsc_comment (surface_, "%%PageOrientation: Landscape");
 
 	  if(magCompare(MAGICS_SITE,"ecmwf"))
-	    cairo_ps_surface_dsc_comment (surface_, "%%Copyright: Copyright (C) ECMWF");
+	    cairo_ps_surface_dsc_comment (surface_, "%%Copyright: ECMWF");
 #endif
 	}
 
-//#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 2, 0)
-//	cairo_surface_set_fallback_resolution (surface_, resolution_, resolution_);
-//#endif
 	if(magCompare(transparent_,"off") || !(magCompare(backend_,"png") || magCompare(backend_,"geotiff")) )
 	{
 		cairo_set_source_rgb (cr_, 1.0, 1.0, 1.0); /* white */
@@ -322,11 +267,6 @@ void CairoDriver::setupNewSurface() const
 	   dimensionY_ =  static_cast<int>(ratio*dimensionXglobal_);
 	}
 	currentPage_ = 0;
-
-	if(magCompare(antialias_,"off"))
-		cairo_set_antialias(cr_, CAIRO_ANTIALIAS_NONE);
-	else
-		cairo_set_antialias(cr_, CAIRO_ANTIALIAS_SUBPIXEL);
 }
 
 /*!
@@ -335,30 +275,13 @@ void CairoDriver::setupNewSurface() const
 void CairoDriver::close()
 {
 	currentPage_ = 0;
+	if(magCompare(backend_,"pdf") && !fileName_.empty()) printOutputName("CAIRO pdf "+fileName_);
+	if(magCompare(backend_,"ps")  && !fileName_.empty()) printOutputName("CAIRO ps "+fileName_);
 
-	if(magCompare(backend_,"pdf") && !filename_.empty()) printOutputName("CAIRO pdf "+filename_);
-	if(magCompare(backend_,"ps") && !filename_.empty()) printOutputName("CAIRO ps "+filename_);
-
-	if ( context_ == 0 ) {
+	if (!context_) {
 		cairo_surface_destroy (surface_);
 		cairo_destroy (cr_);
 	}
-/*
-#if CAIRO_HAS_XLIB_SURFACE
-	if(magCompare(backend_,"x"))
-	{
-		XEvent event;
-		while(1)
-		{
-			XNextEvent(dpy, &event);
-			if(event.type==Expose && event.xexpose.count<1)
-			{}
-			else if(event.type==ButtonPress) break;
-		}
-		XCloseDisplay(dpy);
-	}
-#endif
-*/
 }
 
 
@@ -395,8 +318,9 @@ MAGICS_NO_EXPORT void CairoDriver::startPage() const
 			cairo_destroy (cr_);
 			cairo_surface_destroy (surface_);
 
-			filename_ = getFileName("svg",currentPage_+1);
-			surface_ = cairo_svg_surface_create(filename_.c_str(), dimensionXglobal_, dimensionYglobal_);
+			fileName_ = getFileName("svg",currentPage_+1);
+      MagLog::dev() << "Cairo - SVG - create file " << fileName_ << endl;
+			surface_ = cairo_svg_surface_create(fileName_.c_str(), dimensionXglobal_, dimensionYglobal_);
 			cr_ = cairo_create (surface_);
 #if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 4, 0)
 			cairo_svg_surface_restrict_to_version (surface_, CAIRO_SVG_VERSION_1_1);
@@ -409,8 +333,8 @@ MAGICS_NO_EXPORT void CairoDriver::startPage() const
 			cairo_destroy (cr_);
 			cairo_surface_destroy (surface_);
 
-			filename_ = getFileName("eps",currentPage_+1);
-			surface_ = cairo_ps_surface_create(filename_.c_str(), dimensionXglobal_, dimensionYglobal_);
+			fileName_ = getFileName("eps",currentPage_+1);
+			surface_ = cairo_ps_surface_create(fileName_.c_str(), dimensionXglobal_, dimensionYglobal_);
 			cairo_ps_surface_set_eps (surface_,true);
 			cr_ = cairo_create (surface_);
 #endif
@@ -421,7 +345,11 @@ MAGICS_NO_EXPORT void CairoDriver::startPage() const
 	{
 		cairo_translate(cr_,0,static_cast<MFloat>(dimensionYglobal_));
 	}
-//	cairo_scale(cr_,1,-1);
+
+	if(magCompare(antialias_,"off"))
+		cairo_set_antialias(cr_, CAIRO_ANTIALIAS_NONE);
+	else
+		cairo_set_antialias(cr_, CAIRO_ANTIALIAS_SUBPIXEL);
 
 	currentPage_++;
 	newPage_ = true;
@@ -440,34 +368,40 @@ MAGICS_NO_EXPORT void CairoDriver::endPage() const
 
 	if(magCompare(backend_,"eps"))
 	{
-		if(!filename_.empty()) printOutputName("CAIRO eps "+filename_);
+		if(!fileName_.empty()) printOutputName("CAIRO eps "+fileName_);
 	}
 	else if(magCompare(backend_,"svg"))
 	{
-		if(!filename_.empty()) printOutputName("CAIRO svg "+filename_);
+		if(!fileName_.empty()) printOutputName("CAIRO svg "+fileName_);
 	}
 	else if (magCompare(backend_,"png") )
 	{
 		Timer timer("cairo", "write png");
-		filename_ = getFileName("png" ,currentPage_);
+		fileName_ = getFileName("png" ,currentPage_);
+
+    cairo_status_t status = CAIRO_STATUS_SUCCESS;
 		if(magCompare(palette_,"on"))
 		{
 		   if(!write_8bit_png())
 		   {
-		     MagLog::warning() << "CairoDriver::renderPNG > palletted PNG failed! Generate 24 bit one ..." << endl;
-		     cairo_surface_write_to_png(surface_, filename_.c_str());
+         MagLog::warning() << "CairoDriver::renderPNG > palletted PNG failed! Generate 24 bit one ..." << endl;
+         status = cairo_surface_write_to_png(surface_, fileName_.c_str());
 		   }
 		}
 		else
 		{
-		   cairo_surface_write_to_png(surface_, filename_.c_str());
+		   status = cairo_surface_write_to_png(surface_, fileName_.c_str());
 		}
-		if(!filename_.empty()) printOutputName("CAIRO png "+filename_);
+		if(status != CAIRO_STATUS_SUCCESS){
+      MagLog::error() << "PNG could NOT be written - " << cairo_status_to_string(status) <<endl;
+      MagLog::error() << "^^^^^^^^^^^^^^^^^^^^^^^^" <<endl;
+    }
+		if(!fileName_.empty()) printOutputName("CAIRO png "+fileName_);
 	}
 	else if (magCompare(backend_,"geotiff") )
 	{
 #ifdef HAVE_GEOTIFF
-		filename_ = getFileName("tif" ,currentPage_);
+		fileName_ = getFileName("tif" ,currentPage_);
 		write_tiff();
 #else
 		MagLog::error() << "CairoDriver: GEOTIFF not enabled!"<< std::endl;
@@ -475,6 +409,25 @@ MAGICS_NO_EXPORT void CairoDriver::endPage() const
 	}
 }
 
+
+void CairoDriver::newLayer(Layer&) const
+{
+	cairo_save(cr_);
+}
+
+#include "CairoDriver-blur.h"
+
+void CairoDriver::closeLayer(Layer&) const
+{
+    if(applyGaussianBlur_ > 0)
+    {
+        blur_image_surface(surface_,50);
+    }
+	cairo_restore(cr_);
+}
+
+
+
 #ifdef HAVE_GEOTIFF
 
 #include <geotiffio.h>
@@ -498,16 +451,16 @@ MAGICS_NO_EXPORT void CairoDriver::write_tiff() const
     int          height = cairo_image_surface_get_height(surface_);
     const int    stride = cairo_image_surface_get_stride(surface_);
 
-    TIFF *tif = TIFFOpen(filename_.c_str(), "w");
+    TIFF *tif = TIFFOpen(fileName_.c_str(), "w");
     if (!tif) {
-        MagLog::warning() << "CairoDriver: Unable to open TIFF file "<<filename_<< std::endl;
+        MagLog::warning() << "CairoDriver: Unable to open TIFF file "<<fileName_<< std::endl;
         return;
     }
-    
+
     GTIF *gtif = GTIFNew(tif);
     if (!gtif)
     {
-        MagLog::warning() << "CairoDriver: Unable to open GeoTIFF file "<<filename_<< std::endl;
+        MagLog::warning() << "CairoDriver: Unable to open GeoTIFF file "<<fileName_<< std::endl;
         return;
     }
 
@@ -539,8 +492,8 @@ MAGICS_NO_EXPORT void CairoDriver::write_tiff() const
 
     // DPI
     TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
-    TIFFSetField(tif, TIFFTAG_XRESOLUTION, (float) resolution_);
-    TIFFSetField(tif, TIFFTAG_YRESOLUTION, (float) resolution_);
+    TIFFSetField(tif, TIFFTAG_XRESOLUTION, (float) 90.);
+    TIFFSetField(tif, TIFFTAG_YRESOLUTION, (float) 90.);
 
     unsigned char *buf;
     if (TIFFScanlineSize(tif))
@@ -576,15 +529,16 @@ MAGICS_NO_EXPORT void CairoDriver::write_tiff() const
 
 MAGICS_NO_EXPORT bool CairoDriver::write_8bit_png() const
 {
+/*
     cairo_surface_flush (surface_);
     unsigned char *data = cairo_image_surface_get_data(surface_);
     const int     width = cairo_image_surface_get_width(surface_);
     const int    height = cairo_image_surface_get_height(surface_);
-  
+
     struct pngquant_options options_ = { };
     options_.liq = liq_attr_create();
     struct pngquant_options *options = &options_;
-    //    pngquant_file(filename_.c_str(), filename.c_str(), &options);
+    //    pngquant_file(fileName_.c_str(), filename.c_str(), &options);
 
     pngquant_error  retval             = SUCCESS;
     liq_image*      input_image        = NULL;
@@ -600,12 +554,12 @@ MAGICS_NO_EXPORT bool CairoDriver::write_8bit_png() const
 	data2[h*4*width+w+3] = data[h*4*width+w+3];  // a
       }
     }
- 
+
     input_image = liq_image_create_rgba(options->liq, data2, width, height, 0);
     if (!input_image) {
         //return OUT_OF_MEMORY_ERROR;
     }
-    
+
     int quality_percent = 90; // quality on 0-100 scale, updated upon successful remap
     png8_image output_image = {};
 
@@ -637,24 +591,25 @@ MAGICS_NO_EXPORT bool CairoDriver::write_8bit_png() const
     if (SUCCESS == retval) {
         output_image.fast_compression  = false; //  (fast_compression ? Z_BEST_SPEED : Z_BEST_COMPRESSION);
 
-        retval = write_image(&output_image, NULL, filename_.c_str(), options);
+        retval = write_image(&output_image, NULL, fileName_.c_str(), options);
     }
 
     liq_image_destroy(input_image);
     rwpng_free_image8(&output_image);
 
     if (SUCCESS == retval) return true;
+*/
     return false;
 }
 
-/*  
+/*
 MAGICS_NO_EXPORT void CairoDriver::write_8bit_png() const
 {
-    const string filename = filename_ +"_8bit";    
+    const string filename = fileName_ +"_8bit";
     FILE * fp = fopen (filename.c_str(), "wb");
     if (! fp) {
         MagLog::error() << "CairoDriver: Unable to open 8 bit PNG file "<<filename<< std::endl;
-        return;        
+        return;
     }
 
     cairo_surface_flush (surface_);
@@ -670,7 +625,7 @@ MAGICS_NO_EXPORT void CairoDriver::write_8bit_png() const
     	MagLog::error() << "CairoDriver: Unable to create WRITE struct for 8 bit PNG file "<<filename<< std::endl;
         return;
     }
- 
+
     png_infop info_ptr  = png_create_info_struct (png_ptr);
     if(!png_ptr)
     {
@@ -678,7 +633,7 @@ MAGICS_NO_EXPORT void CairoDriver::write_8bit_png() const
         return;
     }
 //    setjmp (png_jmpbuf (png_ptr));
-    
+
     // Set image attributes
     png_set_IHDR (png_ptr,
                   info_ptr,
@@ -689,21 +644,21 @@ MAGICS_NO_EXPORT void CairoDriver::write_8bit_png() const
                   PNG_INTERLACE_NONE,
                   PNG_COMPRESSION_TYPE_DEFAULT,
                   PNG_FILTER_TYPE_DEFAULT);
-    
+
     // Initialize rows of PNG.
     png_bytep *row_pointers = (png_bytep*) malloc(3 * width * sizeof(png_byte));
 
     for (size_t y = 0; y < height; ++y) {
         row_pointers[y] = data + width * 4 * y;
     }
-    
+
     // Write the image data to file
     png_init_io   (png_ptr, fp);
     png_set_rows  (png_ptr, info_ptr, row_pointers);
     png_write_png (png_ptr, info_ptr, PNG_TRANSFORM_BGR, NULL);
 
- //   free (row_pointers); 
- //   png_destroy_write_struct (&png_ptr, &info_ptr);  
+ //   free (row_pointers);
+ //   png_destroy_write_struct (&png_ptr, &info_ptr);
     fclose (fp);
    return;
 }
@@ -983,7 +938,7 @@ MAGICS_NO_EXPORT void CairoDriver::renderSimplePolygon(const Polyline& line) con
 		vector<double> x;
 		vector<double> y;
 		line.hole(h,x,y);
-		if ( x.empty() ) 
+		if ( x.empty() )
 			continue;
 		cairo_move_to (cr_, projectX(x[0]),setY(projectY(y[0])));
 		vector<double>::const_iterator yt = y.begin();
@@ -1112,7 +1067,7 @@ MAGICS_NO_EXPORT void CairoDriver::renderSimplePolygon() const
 			cairo_move_to(cr2, density+.5*.5, 0);
 			cairo_line_to(cr2, density+.5*.5, density+.5);
 		}
-		if(indexHatch_==4 || indexHatch_==6) 
+		if(indexHatch_==4 || indexHatch_==6)
 		{
 			cairo_move_to(cr2,       0, 0);
 			cairo_line_to(cr2, density+.5, density+.5);
@@ -1168,7 +1123,7 @@ MAGICS_NO_EXPORT void CairoDriver::renderSimplePolygon() const
 
   Cairo expects a string as a char array, where each character is expressed as
   16 bit Unicode. Expat however delivers Multi-Byte encoding!
-  
+
   \sa http://www.pygtk.org/docs/pygtk/pango-markup-language.html
 
   \sa Text
@@ -1188,11 +1143,12 @@ MAGICS_NO_EXPORT void CairoDriver::renderText(const Text& text) const
 	vector<NiceText>::const_iterator niceText = text.textBegin();
 	vector<NiceText>::const_iterator niceTextEnd = text.textEnd();
 
-	MFloat textXoffset = 0.;
+	//MFloat textXoffset = 0.;
 	ostringstream alltext;
 
 	for(;niceText<niceTextEnd;niceText++)
 	{
+		if ((*niceText).text().empty()) continue;
 		MagFont magfont = (*niceText).font();
 		const std::set<string>& styles = magfont.styles();
 
@@ -1285,7 +1241,7 @@ MAGICS_NO_EXPORT void CairoDriver::renderText(const Text& text) const
 
 	  cairo_move_to (cr_, xxx, yyy);
 	  const double angle = text.getAngle();
-	  if(fabs(angle) > epsilon)
+	  if(fabs(angle) > EPSILON)
           {
 	      cairo_rotate (cr_, angle );
 	  }
@@ -1323,9 +1279,12 @@ MAGICS_NO_EXPORT void CairoDriver::circle(const MFloat x, const MFloat y, const
 
 //	cairo_identity_matrix (cr_);
 	cairo_set_line_width (cr_, currentLineWidth_);
-	cairo_set_source_rgba(cr_,currentColour_.red(),currentColour_.green(),currentColour_.blue(),currentColour_.alpha());
-
 	int fill = s;
+	if(fill==10) {
+		cairo_set_line_width (cr_, currentLineWidth_*5);
+		fill=0;
+	}
+	cairo_set_source_rgba(cr_,currentColour_.red(),currentColour_.green(),currentColour_.blue(),currentColour_.alpha());
 
 	//if(s > 8) fill = 8;
 	if( (s > 0) && (fill != 9) )
@@ -1383,7 +1342,7 @@ MAGICS_NO_EXPORT bool CairoDriver::renderPixmap(MFloat x0,MFloat y0,MFloat x1,MF
 	cairo_save(cr_);
 	cairo_antialias_t t = cairo_get_antialias(cr_);
 	cairo_set_antialias(cr_, CAIRO_ANTIALIAS_NONE);
-	
+
 	for(int i=h-1;i>=0;i--)
 	{
 		for(int j=0;j<w; x0+=dx,j++)
@@ -1473,7 +1432,7 @@ MAGICS_NO_EXPORT void CairoDriver::renderImage(const ImportObject& obj) const
 
 
 /*!
-  \brief render cell arrays
+  \brief render cell arrays for cell shading
 
   This method renders cell arrays, also called images in Magics language. These are
   mainly used for satellite data.
@@ -1484,68 +1443,92 @@ MAGICS_NO_EXPORT void CairoDriver::renderImage(const ImportObject& obj) const
 */
 MAGICS_NO_EXPORT bool CairoDriver::renderCellArray(const Image& image) const
 {
-	MagLog::debug() << "CD:renderCellArray> "<<image.getWidth()<<"x"<<image.getHeight() << endl;
-	ColourTable &lt  = image.getColourTable();
-	const int width  = image.getNumberOfColumns();
-	const int height = image.getNumberOfRows();
-	const double tr  = image.getTransparency();
-	const MFloat x0  = projectX(image.getOrigin().x());
-	const MFloat y0  = projectY(image.getOrigin().y());
+  MagLog::debug() << "CD:renderCellArray> "<<image.getWidth()<<"x"<<image.getHeight() << endl;
+  ColourTable &lt  = image.getColourTable();
+  const int width  = image.getNumberOfColumns();
+  const int height = image.getNumberOfRows();
+  //cout << "CD:renderCellArray> "<<width<<"x"<<height << endl;
+  const MFloat x0  = projectX(image.getOrigin().x());
+  const MFloat y0  = projectY(image.getOrigin().y());
+  const MFloat scX = (image.getWidth() *coordRatioX_) /width;
+  const MFloat scY = (image.getHeight()*coordRatioY_) /height;
+  const double wid = projectX(image.getOrigin().x()+image.getWidth())  - projectX(image.getOrigin().x());
+  const double hei = projectY(image.getOrigin().y()+image.getHeight()) - projectY(image.getOrigin().y());
 
-	cairo_save(cr_);
-	cairo_antialias_t t = cairo_get_antialias(cr_);
-	cairo_set_antialias(cr_, CAIRO_ANTIALIAS_NONE);
+  cairo_save(cr_);
+//  cairo_antialias_t t = cairo_get_antialias(cr_);
+//  cairo_set_antialias(cr_, CAIRO_ANTIALIAS_NONE);
+  cairo_translate (cr_, x0, y0);
 
-	const long dim=width*height;
-	unsigned char *chImage = new unsigned char[dim*4];
-	int jj = 0;
-//	const Colour none("none");
-	for(int j=0;j<dim; j++)
-	{
-		const short c = image[j];
+  if( (width > 800) || (height > 600) ) // high resoltion enough to use image
+  {
+    cairo_surface_t *result = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
+    if (cairo_surface_status(result) != CAIRO_STATUS_SUCCESS)
+    {
+      MagLog::warning()  << "CAIRO:renderImage> can not create surface ("<<width<<"x"<<height<<")"<< endl;
+      return result;
+    }
+    cairo_surface_flush(result);
+//  cairo_t* cr_tmp  = cairo_create(surface_);
+//  cairo_set_antialias(cr_tmp, CAIRO_ANTIALIAS_NONE);
 
-		const float cr = lt[c].red();
-		const float cg = lt[c].green();
-		const float cb = lt[c].blue();
-		
-		if(cr*cg*cb <0)
-		{
-			chImage[jj]=0; jj++;
-			chImage[jj]=0; jj++;
-			chImage[jj]=0; jj++;
-			chImage[jj]=0; jj++;
-		}
-		else
-		{
-			const double al = lt[c].alpha();
-			chImage[jj]=char(int(255.*cb *al )); jj++;
-			chImage[jj]=char(int(255.*cg *al )); jj++;
-			chImage[jj]=char(int(255.*cr *al )); jj++;
-			chImage[jj]=char(int(255.*al)); jj++;
-		}
-	}
+    unsigned char *current_row = cairo_image_surface_get_data(result);
+    int stride = cairo_image_surface_get_stride(result);
 
-	cairo_surface_t* surImage = cairo_image_surface_create_for_data(
-		chImage,
-		CAIRO_FORMAT_ARGB32,
-		width,
-		height,
-		width * 4
-	);
+    for(unsigned int h=0;h<height; h++)
+    {
+     uint32_t *row = (uint32_t *)current_row;
+     for(unsigned int w=0;w<width; w++)
+     {
+      const short c  = image[w + (width*h)];
+      double al = lt[c].alpha();
+      if( (lt[c].red()*lt[c].green()*lt[c].blue()<0.) )
+        al=0.;  // missing data will be fully transparent
+      const uint32_t cr = (uint32_t)(al * lt[c].red()   * 255.);
+      const uint32_t cg = (uint32_t)(al * lt[c].green() * 255.);
+      const uint32_t cb = (uint32_t)(al * lt[c].blue()  * 255.);
+      const uint32_t alint = (uint32_t)(al*255.);
+      row[w] =  (alint << 24) | (cr << 16) | (cg << 8) | cb;
+//        row[w] =  (cr << 16) | (cg << 8) | cb;
+     }
+     current_row += stride;
+    }
+    cairo_surface_mark_dirty(result);
 
-	cairo_translate (cr_, x0, y0);
-	const double scX = (image.getWidth() *coordRatioX_) /width;
-	const double scY = (image.getHeight()*coordRatioY_) /height;
-	cairo_scale (cr_, scX, -scY);
+    cairo_scale (cr_, scX, -scY);
+  //  cairo_set_operator(cr_, CAIRO_OPERATOR_OVER);
+    cairo_set_source_surface(cr_, result, 0, 0);
+    cairo_paint(cr_);
 
-	cairo_set_source_surface(cr_, surImage, 0, 0);
-	cairo_paint(cr_);
+    cairo_surface_destroy (result);
+//    cairo_destroy(cr_tmp);
+  }else
+  {
+    cairo_antialias_t t = cairo_get_antialias(cr_);
+    cairo_set_antialias(cr_, CAIRO_ANTIALIAS_NONE);
+    for(unsigned int h=0;h<height; h++)
+    {
+     for(unsigned int w=0;w<width; w++)
+     {
+      const short c  = image[w + (width*h)];
+      const float cr = lt[c].red();
+      const float cg = lt[c].green();
+      const float cb = lt[c].blue();
+      if(cr*cg*cb >=0){
+       cairo_set_source_rgba(cr_,cr,cg,cb,lt[c].alpha());
+       cairo_set_line_width (cr_,0.01);
+       cairo_rectangle (cr_, w*scX, h*-scY, scX, -scY);
+       cairo_fill_preserve(cr_);
+       cairo_stroke(cr_);
+      }
+     }
+    }
+    cairo_set_antialias(cr_, t);
+  }
 
-	cairo_surface_destroy (surImage);
-
-	cairo_restore(cr_);
-	cairo_set_antialias(cr_, t);
-	return true;
+  cairo_restore(cr_);
+//  cairo_set_antialias(cr_, t);
+  return true;
 }
 
 
@@ -1579,33 +1562,39 @@ void CairoDriver::print(ostream& out)  const
 MAGICS_NO_EXPORT void CairoDriver::renderSymbols(const Symbol& symbol) const
 {
 	debugOutput("Start CairoDriver Symbols");
-/*
-	if(symbol.getSymbol()=="logo_ecmwf")
-	{
-		const string logofile = getEnvVariable("MAGPLUS_HOME") + MAGPLUS_PATH_TO_SHARE_ + "ecmwf_logo.png";
-		cairo_surface_t *image = cairo_image_surface_create_from_png(logofile.c_str());
 
-		if(image)
+	const string symbolName = symbol.getSymbol();
+	const string logo = "logo_";
+	if (symbolName.find(logo) == std::string::npos)
+	{
+		BaseDriver::renderSymbols(symbol);
+	}
+	else
+	{
+	    string logofile = getEnvVariable("MAGPLUS_HOME") + MAGPLUS_PATH_TO_SHARE_;
+		if     (symbolName == "logo_cams") logofile += "CAMS_combined.png";
+		else if(symbolName == "logo_c3s")  logofile += "C3S_combined.png";
+		else   logofile += "ecmwf_logo_2014.png";
+
+		cairo_surface_t *image = cairo_image_surface_create_from_png(logofile.c_str());
+		cairo_status_t ret = cairo_surface_status(image);
+
+		if(!ret)
 		{
 			cairo_save(cr_);
-			int w = cairo_image_surface_get_width(image);
-			int h = cairo_image_surface_get_height(image);
-
-			cairo_translate (cr_, projectX(symbol[0].x()), projectY(symbol[0].y()));
-			const MFloat scaling = convertCM(symbol.getHeight()*.5) / coordRatioY_;
-//			cairo_scale (cr_, 0.3, 0.3);
-			cairo_set_source_surface(cr_, image, w*scaling, h*scaling);
+			cairo_translate (cr_, projectX(symbol[0].x()), projectY(symbol[0].y())-convertCM(symbol.getHeight()*.3));
+			const MFloat sizeY = -convertCM(symbol.getHeight() *.1) * coordRatioY_;
+			const int w = cairo_image_surface_get_width(image);
+			const int h = cairo_image_surface_get_height(image);
+            const double ratio = w/h;
+            const MFloat sizeX = sizeY * ratio;
+			cairo_scale (cr_, sizeX/w, sizeY/h);
+			cairo_set_source_surface(cr_, image, 0, 0);
 			cairo_paint(cr_);
-
 			cairo_surface_destroy (image);
 			cairo_restore(cr_);
 		}
-		else MagLog::warning() << "CairoDriver-> Could NOT read the logo file "<< logofile << " !" << endl;
-	}
-//	else if(symbol.getSymbol().compare(0,7,"magics_")==0 )
-	else
-*/	{
-		BaseDriver::renderSymbols(symbol);
+		else MagLog::warning() << "CairoDriver - Could NOT read logo "<< logofile << endl;
 	}
 }
 
